# rrx.rb version 1.0: (coding: euc-jp)  2013/04/15
  # a ruby library for R(statistical analysis command), on various OS
  # Copyright (C) T. Yoshiizumi, 2013 All rights reserved.
require "tempfile"
require "csv"
require "nkf"
if RUBY_PLATFORM =~ /mswin(?!ce)|mingw|cygwin|bccwin/i
  require "win32ole"
end

# sub_methods for Rrx
module RrxSub
  module_function()
  def get_sep(str)  # ڤʸȽ
    cct = tct = 0
    if str.class == String
      str.gsub(/,/) {|ptn|
        cct += 1
        ptn = ''
      }
      str.gsub(/\t/) {|ptn|
        tct += 1
        ptn = ''
      }
    end
    return (tct > cct) ? "\t" : ","
  end

  def rm_upper(ary)  # ζԤ
    ary.map! {|row|
      while row.size > 0 and row.last.to_s =~ /^\s*$/
        row.pop  
      end
      row
    }
    while ary.size > 0 and ary[0] == []
      ary.shift  
    end
    return ary
  end

  def mark_num(ary)  # ޡŤ
    res = []
    for i in 0...ary.size
      row = Array.new
      for j in 0...ary[i].size
        s = ary[i][j].to_s
        if s =~ /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
          row[j] = [i,j]
        else
          row[j] = nil
        end
      end
      res << row
    end
    return res
  end

  def mark_data(ary)  # ǤʤޡŤ
    res = []
    for i in 0...ary.size
      row = Array.new
      for j in 0...ary[i].size
        s = ary[i][j].to_s
        unless s =~ /^\s*$/
          row[j] = [i,j]
        else
          row[j] = nil
        end
      end
      res << row
    end
    return res
  end

  def square_size(ary, addr)
    y1,x1 = addr
    unless ary[y1][x1]
      return [0, nil, nil]
    end
    if !(ary[y1][x1+1] and ary[y1+1] and ary[y1+1][x1] and ary[y1+1][x1+1])
      return [1, [y1], [x1]]
    end
    d = 2
    while ary[y1+d]
      w = ary[y1+d][x1..(x1+d)]
      break  if !w or w.size < (d+1) or w.include?(nil)
      w = []
      y=y1;  x=x1+d
      while y <= (y1+d) and ary[y][x]
        w << ary[y][x]
        y += 1
      end
      break  if !w or w.size < (d+1) or w.include?(nil)
      d += 1
    end
    yy = y1+d-1
    xx = x1+d-1
    y = yy+1
    while ary[y]
      w = ary[y][x1..xx]
      break  if !w or w.size < d or w.include?(nil)
      y += 1
    end
    y2 = y-1
    res1 = [d, (y1..y2).to_a, (x1..xx).to_a]
    res1_v = (y2-y1+1) * (xx-x1+1)
    x = xx+1
    while ary[y1][x]
      w = [ary[y1][x]]
      for y in (y1+1)..yy
        w << ary[y][x]
      end
      break  if !w or w.size < d or w.include?(nil)
      x += 1
    end
    x2 = x-1
    res2 = [d, (y1..yy).to_a, (x1..x2).to_a]
    res2_v = (yy-y1+1) * (x2-x1+1)
    return (res2_v > res1_v) ? res2 : res1
  end

  def grouping_box(boxes)
    res = []
    del_count = 0
    while box = boxes.shift
      d,yy,xx = box
      i = boxes.size-1
      while i >= 0
        d2,yy2,xx2 = boxes[i]
        if (yy & yy2).size > 0 and (xx & xx2).size > 0
          boxes.delete_at(i)
          del_count += 1
          if d2 > d
            d,yy,xx = [d2,yy2,xx2]
          elsif d2 == d and yy2.size*xx2.size > yy.size*xx.size
            d,yy,xx = [d2,yy2,xx2]
          end
        end
        i -= 1
      end
      res << [d,yy,xx]
    end
    if del_count > 0
      res = grouping_box(res)
    end
    return res
  end

  def get_boxes(ary)
    boxes = []
    ary.each_with_index {|row, i|
      next  if row.class != Array  or row.size < 1
      row.each_with_index {|cell, j|
        next  unless cell
        d,yy,xx = square_size(ary, [i,j])
        next  if d < 2
        boxes << [d,yy,xx]
      }
    }
    if boxes.size > 1
      boxes = grouping_box(boxes)
    end
    return boxes
  end

  def each_square(ary, &block)  # ơικɸĤ򤿤ɤ
    unless block
      return
    end
    boxes = get_boxes(ary)
    boxes.each {|box|
      d,yy,xx = box
      w = [yy[0], xx[0], yy[-1], xx[-1]]
      block.call(w)
    }
  end

  def select_num2(ary)  # ĲܸФ
    ary2 = []
    nary = RrxSub.mark_num(ary)
    RrxSub.each_square(nary) {|y1,x1, y2,x2|
      next  if !(y2 > y1 and x2 > x1)
      y0 = (y1>0) ? (y1-1) : y1
      x0 = (x1>0) ? (x1-1) : x1
      read_opt = (y1 > y0) ? "header=T" : "header=F"
      read_opt = read_opt + ", row.names=1"  if x1 > x0
      title = []
      title.push(ary[y0-1][x0-1].to_s)  if y1>y0 and y0>0 and
          x1>x0 and x0>0 and !nary[y0-1][x0-1]
      title.push(ary[y0-1][x0].to_s)  if y1>y0 and y0>0 and !nary[y0-1][x0]
      title.push(ary[y0][x0-1].to_s)  if x1>x0 and x0>0 and !nary[y0][x0-1]
      title.delete_if {|x|  x =~ /^\s*$/}
      wary = []
      for i in y0..y2
        row = Array.new
        for j in x0..x2
          val = ary[i][j]
          if val.to_s =~ /^\s*$/ and [i,j] != [y1-1,x1-1]
            if i == (y1-1)
              val = sprintf("x%d", j-((x1==x0) ? (x0-1) : x0))
            end
            if j == (x1-1)
              val = sprintf("y%d", i-((y1==y0) ? (y0-1) : y0))
            end
          end
          row << val
        end
        wary << row
      end
      ary2 << [wary, title.join("|"), read_opt]
    }
    return ary2
  end

  def select_data2(ary)
    ary2 = []
    sary = RrxSub.mark_data(ary)
    RrxSub.each_square(sary) {|y1,x1, y2,x2|
      next  if !(y2 > y1 and x2 > x1)
      num_regex = /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
      num_flag = false
      for j in x1..x2
        if ary[y1][j].to_s =~ num_regex
          num_flag = true
          break
        end
      end
      read_opt = num_flag ? "header=F" : "header=T"
      num_flag = false
      for i in y1..y2
        if ary[i][x1].to_s =~ num_regex
          num_flag = true
          break
        end
      end
      read_opt += ", row.names=1"  unless num_flag
      title = []
      title.push(ary[y1-1][x1-1].to_s)  if y1>0 and x1>0
      title.push(ary[y1-1][x1].to_s)  if y1>0
      title.push(ary[y1][x1-1].to_s)  if x1>0
      title.delete_if {|x|  x =~ /^\s*$/}
      wary = []
      for i in y1..y2
        row = Array.new
        for j in x1..x2
          row << ary[i][j]
        end
        wary << row
      end
      ary2 << [wary, title.join("|"), read_opt]
    }
    return ary2
  end

  def select_squares(ary, type)  # Ĳ
    ary2 = []
    nary = (type == :num) ? RrxSub.mark_num(ary) : RrxSub.mark_data(ary)
    RrxSub.each_square(nary) {|y1,x1, y2,x2|
      if y2 > y1 and x2 > x1
        wary = []
        for i in y1..y2
          row = Array.new
          for j in x1..x2
            row << ary[i][j]
          end
          wary << row
        end
        ary2 << wary
      end
    }
    return ary2
  end

  def table_turn(aa)  # ġž֤
    unless aa.class == Array
      return false
    end
    if aa[0].class != Array
      aa = [aa]
    end
    i_size = aa.size-1
    j_size = 0
    for i in 0..i_size
      next  unless aa[i].class == Array
      if (aa[i].size-1) > j_size
        j_size = aa[i].size-1
      end
    end
    if j_size == 0
      return false
    end
    r = []
    for j in 0..j_size
      w = []
      for i in 0..i_size
        w << aa[i][j]
      end
      r << w
    end
    return r
  end

  def to_num(x)
    s = x.to_s
    if s =~ /^-?\d+$/
      x = s.to_i
    elsif s =~ /^(-?\d+)\.0+$/
      x = $1.to_i
    elsif s =~ /^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
      x = s.to_f
    end
    return x
  end

  def temp_prefix()
    ff = Tempfile.new("ww")
    ff.close
    res = File.dirname(ff.path)
    ff.unlink
    return res
  end
end

# main methods
module Rrx
  @@rrx_cfg = sprintf("%s/rrx.cfg", File.dirname(
      defined?(ExerbRuntime) ? ExerbRuntime.filepath :
      (ENV['OCRA_EXECUTABLE'] ? ENV['OCRA_EXECUTABLE'] : __FILE__)))
  @@path = nil  # RΥեѥ̾
  @@opt = nil
  @@log = nil
  @@encopt = true
  @@prefix = "rrx_"  # ƥݥե̾Ƭ
  @@file_hs = {}
  @@temp_hs = {}
  @@temp_prefix = RrxSub.temp_prefix()
  @@ext_regex = /\.(xls|xml|xlsx|xlsm|html?|ods)$/i
  @@charset = nil

  NCNV = {'e'=>"-e", 'euc-jp'=>"-e", 's'=>"--sjis", 'shift_jis'=>"-s",
      'u'=>"-w", 'utf-8'=>"-w", 'windows-31j'=>"--sjis", 'cp932'=>"--sjis"}
  FS = "\001\001"
  RS = "\002\002"
  VFS = "\005\005"
  CSV_MARK = "!!csv"
  @@robj = "robj.count <- 0\n" +
    "robj.out <- function(item, tt, outfile) {\n" +
    "\tcat(item, \"#{FS}\", file=outfile, sep=\"\")\n" +
    "\tif (!(is.null(dim(tt)))) {\n" +
    "\t\tcat(\"!!csv#{FS}\", file=outfile, sep=\"\")\n" +
    "\t\twrite.table(tt, file=outfile, sep=\",\", quote=T, col.names=NA)}\n" +
    "\telse cat(tt, file=outfile, sep=\"#{VFS}\")\n" +
    "\tcat(\"#{RS}\", file=outfile, sep=\"\")\n}\n" +
    "robj <- function(xx, kk=NULL) {\n" +
    "\tif (is.null(kk)) {kk <- robj.count + 1; robj.count <<- kk;\n" +
        "\t\tfname <- sprintf(\"#{@@temp_prefix}/ROBJ_robj%.0f.rrx\",kk)}\n" +
    "\telse  {fname <- sprintf(\"#{@@temp_prefix}" +
        "/ROBJ_%s.rrx\",kk)}\n" +
    "\toutfile <- file(fname, open=\"w\")\n" +
    "\tnn <- names(xx)\n" +
    "\tif (is.null(nn) || !is.null(dim(xx))) {" +
        "robj.out(\"self\", xx, outfile)\n" +
    "\t\treturn(invisible(xx))}\n" +
    "\tfor (i in 1:length(nn)) {\n" +
    "\t\titem <- nn[i]\n" +
    "\t\ttt <- xx[[item]]\n" +
    "\t\trobj.out(item, tt, outfile)\n" +
    "\t}\n\tclose(outfile)\n}\n"

  module_function()
  def path=(val)
    old_path = @@path
    if val.class == String
      @@path = val
      unless old_path
        hs = rload()
        hs[:path] = @@path
        rsave(hs)
      end
    end
    @@path
  end
  def path()
    if !@@path and test(?f, @@rrx_cfg)
      hs = rload()
      @@path = hs[:path]
    end
    if !@@path
      wpath = search_r_path()
      @@path = !wpath ? "R" : wpath
    end
    @@path
  end
  def cfg_delete()
    if test(?e, @@rrx_cfg)
      File.unlink(@@rrx_cfg)
    end
  end
  def opt=(val)
    @@opt = val
  end
  def opt()
    @@opt
  end
  def log=(val)
    @@log = val
  end
  def log()
    @@log
  end
  def encopt=(val)
    @@encopt = val
  end
  def encopt()
    @@encopt
  end

  def rencode(body)
    if body.class != String or body == ''
      return body
    end
    if @@charset != nil
      code = @@charset
      c = NCNV[code.downcase]
      if c != nil
        body = NKF.nkf("-m0 #{c}", body)
        if RUBY_VERSION.to_f >= 1.9
          body.force_encoding(code)
        end
      end
    elsif RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      if c != "" and NCNV[c]
        c = NCNV[c]
        body = NKF.nkf("-m0 #{c}", body)
      end
    else
      etype = 'a'.encoding.to_s  # ץȤʸ
      c = NCNV[etype.downcase]
      body = NKF.nkf("-m0 #{c}", body)  if c != nil
      body.force_encoding(etype)
    end
    return body
  end

  def file_read(fname)
    mode = (RUBY_VERSION < "1.9") ? "r" : "r:ASCII-8BIT"
    begin
      body = File.open(fname, mode) {|ff|  ff.read}
    rescue
      body = ''
    end
    return rencode(body)
  end

  def rexec(script, logname=nil, options=nil)  # Rץμ¹
    rload(:set)
    @@charset = (script.class == String) ? charset(script) : nil
    res = {}
    if script !~ /\r|\n/ and test(?f, script)
      script = file_read(script)
    end
    if !script or script == ''
      STDERR.puts "the R program is empty."
      return res
    end
    Dir.glob("#{@@temp_prefix}/ROBJ_*.rrx") {|fname|  File.unlink(fname)}
    gv = global_variables.dup
    %w($stdin $stderr $stdout $defout $deferr).each {|key| gv.delete(key)}
    script2 = script.dup
    while script2 =~ /\@\@\$[a-z].*?\@\@/
      script2 = script2.gsub(/\@\@(\$[a-z].*?)\@\@/) {|ptn|
        kk = $1.dup
        if (vv = eval(kk)) != nil
          ptn = "#{vv}"
        else
          STDERR.puts "can't replace '@@#{kk}@@' in method 'rexec'"
          ptn = "??"
        end
      }
    end
    tf = []
    script2 = script2.gsub(/\@\@(.+?)\@\@/) {|ptn|
      key = $1.dup
      unless @@file_hs[key]
        ext = nil
        if key =~ /^.+(\.[a-z0-9]+)$/i
          ext = $1.dup
        end
        tt = @@file_hs[key] = temp_make(ext, false)
        tf << tt
      end
      ptn = @@file_hs[key].path
    }
    temp_delete(@@file_hs[:sink])  if @@file_hs[:sink]
    tt = @@file_hs[:sink] = temp_make(false)
    tf << tt
    prescript = "#{@@robj}"
    if @@encopt and @@charset != nil
      cset = (@@charset =~ /Windows-31J/i) ? 'cp932' : @@charset
      prescript = prescript + "options(encoding=\"#{cset}\")\n"
    end
    prescript = prescript + "sink(file=\"#{@@file_hs[:sink].path}\", " +
        "split=T) ## rrx_sink\n"
    script2 = prescript + script2
    temp_delete(@@file_hs[:script])  if @@file_hs[:script]
    tt = @@file_hs[:script] = temp_make(script2, false)
    tf << tt
    lff = temp_make(false)
    tf << lff
    if options != nil
      opt = options
    else
      opt = !@@opt ? "--no-save -q" : @@opt
    end
    unless opt =~ /--save|--no-save|--vanilla/i
      opt = opt + " --no-save"
    end
    if @@encopt and opt !~ /--encoding/i and @@charset != nil
      cset = (@@charset =~ /Windows-31J/i) ? 'cp932' : @@charset
      opt = opt + " --encoding=#{cset}"
    end
    cmd = "\"#{path()}\" CMD BATCH #{opt} " +
        "\"#{@@file_hs[:script].path}\" \"#{lff.path}\""
    sres = system( cmd )
    logname = @@log  unless logname
    unless sres
      cmd2 = "\"#{path()}\" BATCH #{opt} R_program"
      cmd2 += " \"#{logname}\""  if logname
      STDERR.print "** failed to execute the command\n\t'#{cmd2}'\n\n"
    end
    if test(?z, lff.path)
      tf.delete(lff)
      temp_delete(lff)
    elsif test(?f, lff.path)
      body = rencode( temp_read(lff) )
      body = body.sub(/\A(.+? ## rrx_sink)\n(>.+)\Z/m, "\\2")
      res[:log] = body
      tf.delete(lff)
      temp_delete(lff)
      if logname
        File.open(logname, "w") {|ff|  ff.print body}
      end
      if !sres
        STDERR.print "** log:\n" + res[:log] + "\n"
      end
    end
    @@file_hs.each {|key, ff|
      next  if !test(?f, ff.path) or test(?z, ff.path)
      res[key] = rencode( temp_read(ff) )
      if key == :sink
        res[key] = res[key].sub(/^> proc\.time\([^>]+\Z/, "")
        res[key] = res[key].gsub(/ +\n/, "\n")
      end
    }
    while tt = tf.shift
      temp_delete(tt)
    end
    Dir.glob("#{@@temp_prefix}/ROBJ_*.rrx") {|fname|
      key = nil
      if fname =~ /[\/\\]ROBJ_(.+)\.rrx$/  then  key=$1;  end
      unless key
        File.unlink(fname)
        next
      end
      val = file_read(fname)
      File.unlink(fname)
      res[key] = robj_convert(val)
    }
    return res
  end

  def ary2num(ary)  # ǤοͲ
    res = []
    if ary[0].class != Array
      ary.each {|x|  res.push(RrxSub.to_num(x))}
    else
      ary.each {|row|
        w = []
        row.each {|x|  w.push(RrxSub.to_num(x))}
        res << w
      }
    end
    return res
  end

  def str2ary(str, sep=nil)  # ʸѴ
    sep = RrxSub.get_sep(str)  if sep == nil
    ary = (RUBY_VERSION < "1.9") ? CSV.parse(str, sep) :
        CSV.parse(str, {:col_sep=>sep})
    ary.map! {|row|
      row.map! {|x|
        x = x.to_s
        if x =~ /^\s*$/
          x = nil
        end
        x
      }
      row.pop  while row.size>0 and row.last==nil
      row
    }
    ary = ary2num(ary)
    return ary
  end

  def ary2str(aa, sep=',')  # ʸѴ
    str = ''
    aa.map! {|row|
      row.map! {|x|
        if x.to_s =~ /^\s*$/
          x = nil
        end
        x
      }
    }
    aa.each {|row|
      row = [row]  if row.class != Array
      line = (RUBY_VERSION < "1.9") ? CSV.generate_line(row, sep) :
          CSV.generate_line(row, {:col_sep=>sep})
      line.chomp!
      str += "#{line}\n"
    }
    return str
  end

  def temp_make(*arg)  # ƥݥե
    last_flag =  (arg.size>0 and [true,false].include?(arg.last)) ?
        arg.pop : true
    str = ext = nil
    arg.each {|x|
      if x.class == Symbol or (x.class == String and
          x =~ /^\.[a-z0-9]+$/i)
        if ext == nil
          ext = x.to_s
          ext = '.' + ext  unless ext =~ /^\./
        end
      elsif x.class == String or x.class == Array
        str = x  if str == nil
      end
    }
    if str.class == Array
      sep = (ext.to_s.downcase == '.txt') ? "\t" : ","
      str = ary2str(str, sep)
    end
    w = @@prefix
    w = [w, ext]  if ext
    ff = Tempfile.new(w)
    ff.write str  if str != nil
    ff.close
    @@temp_hs[ff.path] = ff
    @@file_hs['last'] = ff  if last_flag == true
    return ff
  end

  def temp_read(ff)  # ƥݥեƤʸ֤
    ff.open
    str = ff.read
    ff.close
    return str
  end

  def temp_write(ff, str)  # ƥݥե˽񤭹
    res = nil
    if ff and str
      ff.open
      res = ff.write str
      ff.close
    end
    return res
  end

  def temp_delete(ff = nil)  # ƥݥեκ
    if ff != nil
      key = (RUBY_VERSION < "1.9") ? @@temp_hs.index(ff) : @@temp_hs.key(ff)
      @@temp_hs.delete(key)  if key and @@temp_hs[key]
      key = (RUBY_VERSION < "1.9") ? @@file_hs.index(ff) : @@file_hs.key(ff)
      @@file_hs.delete(key)  if key and @@file_hs[key]
      ff.close
      ff.unlink
      return
    end
    if @@file_hs.size > 0
      @@file_hs.each {|key, ff|
        ff.close
        ff.unlink
      }
      @@file_hs.clear
    end
    if @@temp_hs.size > 0
      @@temp_hs.each {|fname, ff|
        next  unless test(?e, fname)
        ff.close
        ff.unlink
      }
      @@temp_hs.clear
    end
  end

  def temp_delete_all()  # Ϣƥݥե
     temp_delete()
    ff = temp_make(false)
    dd = File.dirname(ff.path)
    temp_delete(ff)
    dd = dd.gsub(/\\/, "/")
    dd.sub(/\/$/, "")
    Dir.glob("#{dd}/#{@@prefix}*") {|fname|
      File.unlink(fname)
    }
  end

  def temp_file(key)  # keyбTempfile֥Ȥ֤
    ff = nil
    if key
      ff = @@file_hs[key]
      unless ff
        ff = @@temp_hs[key]
      end
    end
    return ff
  end

  def temp_key()  # Tempfilekey֤
    res = []
    fa = []
    @@file_hs.each {|key, ff|
      res << key
      fa << ff
    }
    @@temp_hs.each {|key, ff|
      next  if fa.include?(ff)
      res << key
    }
    return res
  end

  def temp_keyset(ff, key)
    @@file_hs[key] = ff
  end

  def rsave(hs = nil)  # rrx.cfg¸
    if hs.class == Hash and hs.size > 0
      str = ''
      hs.each {|key, val|
        case key
        when :path
          str = str + "path\t#{val}\n"  if test(?e, val)
        when :opt
          str = str + "opt\t#{val}\n"
        when :log
          str = str + "log\t#{val}\n"
        when :encopt
          str = str + "encopt\t#{val}\n"
        end
      }
      if str != ''
        File.open(@@rrx_cfg, "w") {|ff|  ff.write str}
      end
    else
      File.open(@@rrx_cfg, "w") {|ff|
        ff.print "path\t#{@@path}\n" +
            "opt\t#{@@opt}\n" +
            "log\t#{@@log}\n" +
            "encopt\t#{@@encopt}\n"
      }
    end
  end

  def rload(sw = nil)  # rrx.cfgɹ
    hs = {}
    if !test(?f, @@rrx_cfg) or (str = file_read(@@rrx_cfg)) == ''
      return hs
    end
    str.split(/\n/).each {|line|
      line = line.sub(/^\s+/, "").sub(/\s+$/, "")
      next  unless line =~ /\t/
      a = line.split(/\t/)
      case a[0]
      when 'path'
        wpath = (a[1] =~ /^\s*$/) ? nil : a[1]
        hs[:path] = wpath  if wpath
      when 'opt'
        v = (a[1] =~ /^\s*$/) ? nil : a[1]
        hs[:opt] = v  if v
      when 'log'
        hs[:log] = (a[1] =~ /^\s*$/) ? nil : a[1]
      when 'encopt'
        if a[1] =~ /true/i
          hs[:encopt] = true
        elsif a[1] =~ /false/i
          hs[:encopt] = false
        else
          hs[:encopt] = nil
        end
      end
    }
    if sw == :set
      hs.each {|key, val|
        case key
        when :path  then  @@path=val if @@path == nil;
        when :opt  then  @@opt=val if @@opt == nil;
        when :log  then  @@log=val if @@log == nil;
        when :encopt  then  @@encopt=val if @@encopt == nil;
        end
      }
    end
    return hs
  end

  def delmargin(ary)  # ι
    aa = []
    ary.each {|row|  aa.push(row.dup)}
    mr = aa.size-1
    mc = aa[0].size-1
    if aa[0][0].to_s =~ /^-?\d+$|^-?\d+\.\d+$|^-?\d+\.\d+[Ee](-|\+)*\d+$/
      rs=0;  cs=0
    else
      rs=1;  cs=1
    end
    flag = true
    for i in rs..mr
      unless (aa[i].size-1) == mc
        flag = false
        break
      end
      ttl = 0.0
      for j in cs...mc  # üַספĴ٤
        ttl += aa[i][j].to_f
      end
      unless ttl == aa[i][mc].to_f
        flag = false
        break
      end
    end
    if flag
      for i in 0..mr
        aa[i].delete_at(mc)
      end
      return aa
    end
    flag = true
    for i in rs..mr
      unless (aa[i].size-1) == mc
        flag = false
        break
      end
      ttl = 0.0
      for j in (cs+1)..mc  # üַספĴ٤
        ttl += aa[i][j].to_f
      end
      unless ttl == aa[i][cs].to_f
        flag = false
        break
      end
    end
    if flag
      for i in 0..mr
        aa[i].delete_at(cs)
      end
    end
    return aa
  end

  def del_sum(aa)  # ġξι
    flag = :array
    sep = nil
    if aa.class == String
      sep = RrxSub.get_sep(aa)
      aa = str2ary(aa, sep)
      flag = :string
    end
    aa = delmargin(aa)
    aa = RrxSub.table_turn(aa)
    aa = delmargin(aa)
    aa = RrxSub.table_turn(aa)
    if flag == :string
      res = ary2str(res, sep)
    end
    return aa
  end

  def cut_num(*args, &block)  # (Ĳ)С
    res = []
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_squares(ary, :num)
      arys.each {|wary|
        w = del_sum(wary)
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        block.call(w)  if block
        res << w
      }
    end
    return res
  end

  def cut_num2(*args, &block)  # ɽС
    res = []
    rcv = 1
    if [1,2,3].include?(args.last)
      rcv = args.pop
    end
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_num2(ary)
      arys.each {|wary, title, read_opt|
        w = del_sum(wary)
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        if block
          if rcv == 1
            block.call(w)
            res << w
          elsif rcv == 2
            block.call(w, title)
            res << [w, title]
          elsif rcv == 3
            block.call(w, title, read_opt)
            res << [w, title, read_opt]
          end
        end
      }
    end
    return res
  end

  def cut_data(*args, &block)  # 
    res = []
    rcv = 1
    if [1,2,3].include?(args.last)
      rcv = args.pop
    end
    while arg = args.shift
      if arg.class == String and arg !~ /\r|\n/ and arg =~ @@ext_regex
        ww = xls2ary(arg)
        if ww.size > 0
          while w = ww.pop
            args.unshift(w)
          end
        else
          STDERR.puts "can't open or read the file '#{arg}'"
        end
        next
      end
      flag = nil
      if arg.class == String and arg !~ /\r|\n/
        if  !test(?f, arg) or (body = file_read(arg)) == ''
          STDERR.puts "can't open the file '#{arg}'"
          next
        else
          flag = :file
        end
      else
        body = arg
      end
      if body.class == String
        sep = RrxSub.get_sep(body)
        ary = str2ary(body)
        flag = :string
      elsif body.class == Array
        ary = body
        flag = :array
      end
      if flag == nil
        next
      end
      arys = RrxSub.select_data2(ary)
      arys.each {|w, title, read_opt|
        if flag == :string or flag == :file
          w = ary2str(w, sep)
        end
        if block
          if rcv == 1
            block.call(w)
            res << w
          elsif rcv == 2
            block.call(w, title)
            res << [w, title]
          elsif rcv == 3
            block.call(w, title, read_opt)
            res << [w, title, read_opt]
          end
        end
      }
    end
    return res
  end

  def select_cells(ary, condition, flag=nil, &block)  # 
    res = Array.new
    col = ary[0]
    for i in 1...ary.size
      row = ary[i]
      for j in 1...row.size
        value = row[j]
        if eval(condition)
          if flag == :addr or flag == :address
            res << [row[0], col[j], [i,j]]
          else
            res << [row[0], col[j], value]
          end
          block.call(res.last)  if block
        end
      end
    end
    return res
  end

  def table_turn(aa)
    RrxSub.table_turn(aa)
  end

  def del_row(ary, *nums)  # λԤκ
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    item_w = RrxSub.table_turn(ary)[0]
    na = []
    nums.map! {|x|
      if x.class == String
        item_w.each_with_index {|s, n|
          if s == x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Regexp
        item_w.each_with_index {|s, n|
          if s =~ x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      end
      x
    }
    nums.delete(nil)
    nums = nums + na
    res = []
    ary.each_with_index {|row, i|
      next  if nums.include?(i)
      res << row.dup
    }
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def del_col(ary, *nums)  # λκ
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    item_w = ary[0]
    na = []
    nums.map! {|x|
      if x.class == String
        item_w.each_with_index {|s, n|
          if s == x
            na << n    unless nums.include?(n)
          end
        }
        x = nil
      elsif x.class == Regexp
        item_w.each_with_index {|s, n|
          if s =~ x
            na << n  unless nums.include?(n)
          end
        }
        x = nil
      end
      x
    }
    nums.delete(nil)
    nums = nums + na
    aa = RrxSub.table_turn(ary)
    res = []
    aa.each_with_index {|col, i|
      next  if nums.include?(i)
      res << col
    }
    res = RrxSub.table_turn(res)
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def ins_row(ary, n, *arg)
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    res = []
    ary.each_with_index {|row, i|
      if i == n
        arg.each {|x|  res << x.dup}
      end
      res << row.dup
    }
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def ins_col(ary, n, *arg)
    flag = :array
    sep = nil
    if ary.class == String
      sep = RrxSub.get_sep(ary)
      ary = str2ary(ary, sep)
      flag = :string
    end
    aa = RrxSub.table_turn(ary)
    res = []
    aa.each_with_index {|row, i|
      if i == n
        arg.each {|x|  res << x.dup}
      end
      res << row
    }
    res = RrxSub.table_turn(res)
    if flag == :string
      res = ary2str(res, sep)
    end
    return res
  end

  def xls2ary(filename)
    res = []
    ext = File.extname(filename).downcase
    if !test(?e, filename) or
        !%w(.xls .xml .xlsx .xlsm .htm .html .ods).include?(ext)
      return res
    end
    if defined?(Spreadsheet) and ext == ".xls"
      book = Spreadsheet.open(filename, "rb")
      book.worksheets.each do |ss|
        ary = []
        ss.each do |row|
          row2 = []
          row.each do |cell|
            row2 << cell
          end
          ary << row2
        end
        ary = str2ary( rencode( ary2str(ary) ) )
        res << ary
      end
    end
    return res
  end

  def robj_convert(str)
    res = {}
    str.split(RS).each do |xx|
      a = xx.split(FS)
      if a[1] == CSV_MARK
        key, dummy, val = a
        res[key] = str2ary(val, ",")
      else
        key, val = a
        if val.include?(VFS)
          val = val.split(VFS)
          val.delete("")
          val.delete(nil)
          val = ary2num(val)
        else
          val = RrxSub.to_num(val)
        end
        res[key] = val
      end
    end
    return res
  end

  def charset(smp = nil)
    res = nil
    if RUBY_VERSION < "1.9"
      c = $KCODE.to_s.downcase[0,1]
      hs = {'e'=>'euc-jp', 's'=>'cp932', 'u'=>'utf-8'}
      res = hs[c]
    else
      res = !smp ? 'a'.encoding.to_s : smp.encoding.to_s
    end
    return res
  end

  def search_r_path()
    res = nil
    unless defined?(WIN32OLE)
      return res
    end
    whome = WIN32OLE.new("WScript.Shell").SpecialFolders("MyDocuments")
    whome = whome.sub(/^[A-Z]\:/i, "").gsub(/\\/, "/") + "/R"
    dir_list = ['/R', '/Program Files/R', '/Program Files (x86)/R', whome]
    file_system = WIN32OLE.new("Scripting.FileSystemObject")
    drives = file_system.Drives
    drives.each do |drive|
      next  unless drive.IsReady
      dl = "#{drive.DriveLetter}"
      base_dir = nil
      dir_list.each do |dd|
        w = "#{dl}:#{dd}"
        if test(?d, w)
          base_dir = w
          break
        end
      end
      next  unless base_dir
      ary = []
      Dir.glob("#{base_dir}/R-*") {|dname|
        next  unless test(?d, dname)
        fname = "#{dname}/bin/R.exe"
        if test(?e, fname)
          ary << fname
        end
      }
      if ary.size > 0
        if ary.size > 1
          nl = '000.000'
          vl = nil
          ary.each {|val|
            wval = val.sub(/^.+\/R-(\d.+?)\/bin\/R\.exe$/i, "\\1")
            nary = wval.split(/[^0-9]+/)
            nary.map! {|n|
              n = sprintf("%03d", n.to_i)
            }
            ns = nary.join(".")
            if ns > nl
              nl = ns
              vl = val
            end
          }
          res = vl
        else
          res = ary[0]
        end
        break
      end
    end
    return res
  end

  alias :a2n :ary2num
  alias :s2a :str2ary
  alias :a2s :ary2str
  alias :cut_table :cut_num2
  alias :del_yoko :del_row
  alias :del_tate :del_col
  alias :del_horizontal :del_row
  alias :del_vertical :del_col
  alias :ins_yoko :ins_row
  alias :ins_tate :ins_col
  alias :ins_horizontal :ins_row
  alias :ins_vertical :ins_col
  alias :x2a :xls2ary
  module_function(:a2n, :s2a, :a2s, :x2a, :cut_table)
  module_function(:del_yoko, :del_tate, :del_horizontal, :del_vertical)
  module_function(:ins_yoko, :ins_tate, :ins_horizontal, :ins_vertical)
end
